/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.gpdv.core.api.abstractgraph;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;

/**
 *
 * @author odi
 */
public class AbstractGraph implements GraphPropertyProvider, AbstractEdgeChangeListener, AbstractGraphNodeChangeListener, Lookup.Provider{
    private final String name;
    private final long id;
    private final List<AbstractGraphNode> nodes;
    private final List<AbstractGraphEdge> edges;
    private final List<AbstractGraphChangeListener> listener;
    private final List<AbstractGraphNodeChangeListener> nodeListener;
    private final List<AbstractEdgeChangeListener> edgeListener;
    private Map<String, String> properties;
    private PropertyChangeSupport changeSupport;
    private final Lookup lookup;

    public AbstractGraph(final long graphID, final String name, Object object) throws IllegalArgumentException{
        if(name==null) throw new IllegalArgumentException();
        this.name = name;
        this.id = graphID;
        this.nodes = new ArrayList<AbstractGraphNode>();
        this.edges = new ArrayList<AbstractGraphEdge>();
        this.listener = new ArrayList<AbstractGraphChangeListener>();
        this.edgeListener = new ArrayList<AbstractEdgeChangeListener>();
        this.nodeListener = new ArrayList<AbstractGraphNodeChangeListener>();
        this.properties = new HashMap<String, String>();
        this.changeSupport = new PropertyChangeSupport(this);
        if(object!=null){
            this.lookup = Lookups.fixed(object);
        } else {
            this.lookup = Lookup.EMPTY;
        }
    }

    public final String getName() {
        return name;
    }

    public final long getID() {
        return id;
    }

    public final List<AbstractGraphNode> getNodes() {
        ArrayList<AbstractGraphNode> l = new ArrayList<AbstractGraphNode>();
        l.addAll(nodes);
        return Collections.unmodifiableList(l);
    }

    public final List<AbstractGraphEdge> getEdges() {
        ArrayList<AbstractGraphEdge> l = new ArrayList<AbstractGraphEdge>();
        l.addAll(edges);
        return Collections.unmodifiableList(l);
    }

    /**
     * Function to insert an AbstractGraphNode into the graph
     * The AbstractGraphNode must meet the following criteria to be inserted:
     *      node is not null
     *      node.getGraph() returns the graph the function is called on
     *      the node is not in the list of nodes before the insert
     * @param node the node to insert into the graph
     * @return true if the node was inserted, false otherwise
     */
    public final boolean addNode(final AbstractGraphNode node) {
        if(node==null||node.isInserted()||node.getGraph()!=this) return false;
        beforeAddNode(node);
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.NODE_ADDED, node, null, null);
        nodes.add(node);
        node.addAbstractNodeChangeListener(this);
        node.inserted();
        this.fireAbstractGraphChangeEvent(e);
        afterAddNode(node);
        return true;
    }

    /**
     * Function to insert an AbstractGraphEdge into the graph
     * The AbstractGraphEdge must meet the following criteria to be inserted:
     *      edge is not null
     *      edge.getGraph() returns the graph the function is called on
     *      the edge is not in the list of edges before the insert
     *      the source and destination nodes of edge are in the list of nodes
     * @param edge the edge to insert into the graph
     * @return true if the edge was inserted, false otherwise
     */
    public final boolean addEdge(final AbstractGraphEdge edge) {
        if(edge==null||edge.isInserted()||edge.getGraph()!=this) return false;
        if(!edge.getSource().isInserted()||!edge.getDestination().isInserted()) return false;
        beforeAddEdge(edge);
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.EDGE_ADDED, null, edge, null);
        edges.add(edge);
        edge.getSource().addOutgoingEdge(edge);
        edge.getDestination().addIncomingEdge(edge);
        edge.addAbstractEdgeChangeListener(this);
        edge.inserted();
        this.fireAbstractGraphChangeEvent(e);
        afterAddEdge(edge);
        return true;
    }

    /**
     * Function to remove an AbstractGraphNode from the graph
     * The AbstractGraphNode must meet the following criteria to be removed successfully:
     *      node is not null
     *      node.getGraph() returns the graph the function is called on
     *      the node is in the graphs list of nodes
     * Calling this function also removes the nodes incoming and outgoing edges from the graph!
     * @param node the node to be removed
     * @return true if the node was removed, false otherwise
     */
    public final boolean removeNode(final AbstractGraphNode node) {
        if(node==null||node.getGraph()!=this||!node.isInserted()) return false;
        beforeRemoveNode(node);
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.NODE_REMOVED, node, null, null);
        // remove edges attached to node
        List<AbstractGraphEdge> nEdges = node.getIncomingEdges();
        for(AbstractGraphEdge edge : nEdges){
            this.removeEdge(edge);          // remove incoming edges from graph
        }
        nEdges = node.getOutgoingEdges();
        for(AbstractGraphEdge edge : nEdges){
            this.removeEdge(edge);          // remove outgoing edges from graph
        }
        nodes.remove(node);
        node.removeAbstractNodeChangeListener(this);
        node.removed();
        this.fireAbstractGraphChangeEvent(e);
        afterRemoveNode(node);
        return true;
    }

    public final boolean removeEdge(final AbstractGraphEdge edge) {
        if(edge==null||edge.getGraph()!=this||!edge.isInserted()) return false;
        beforeRemoveEdge(edge);
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.EDGE_REMOVED, null, edge, null);
        if(edge.getSource()!=null) edge.getSource().removeOutgoingEdge(edge);          // remove from source node
        if(edge.getDestination()!=null) edge.getDestination().removeIncomingEdge(edge);     // remove from destination node
        edges.remove(edge);
        edge.removeAbstractEdgeChangeListener(this);
        edge.removed();
        this.fireAbstractGraphChangeEvent(e);
        afterRemoveEdge(edge);
        return true;
    }

    public final synchronized void addAbstractGraphChangeListener(AbstractGraphChangeListener listener) {
        this.listener.add(listener);
    }

    public final synchronized void removeAbstractGraphChangeListener(AbstractGraphChangeListener listener) {
        this.listener.remove(listener);
    }

    public final Map<String, String> getProperties() {
        return Collections.unmodifiableMap(properties);
    }

    public final void addProperty(final String key, final String value) {
        if(key==null) return;
        beforeAddProperty(key, value);
        Map<String, String> map = new HashMap<String, String>();
        map.put(key, value);
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.PROPERTY_ADDED, null, null, map);
        properties.put(key, value);
        changeSupport.firePropertyChange(key, null, value);
        this.fireAbstractGraphChangeEvent(e);
        afterAddProperty(key, value);
    }

    public final void removeProperty(final String key) {
        if(key==null||!properties.containsKey(key)) return;
        beforeRemoveProperty(key);
        Map<String, String> map = new HashMap<String, String>();
        map.put(key, properties.get(key));
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.PROPERTY_REMOVED, null, null, map);
        String s = properties.get(key);
        properties.remove(key);
        changeSupport.firePropertyChange(key, s, null);
        this.fireAbstractGraphChangeEvent(e);
        afterRemoveProperty(key);
    }

    public final void setProperties(final Map<String, String> properties) {
        beforeSetProperties(properties);
        Map<String, String> map = new HashMap<String, String>(this.properties);
        AbstractGraphChangeEvent e = new AbstractGraphChangeEvent(this, AbstractGraphChangeEvent.NEW_PROPERTIES, null, null, map);
        if(properties==null){
            for(String s : this.properties.keySet()){
                changeSupport.firePropertyChange(s, this.properties.get(s), null);
            }
            this.properties.clear();
        } else {
            this.properties = properties;
            for(String s : this.properties.keySet()){
                changeSupport.firePropertyChange(s, this.properties.get(s), null);
            }
        }
        this.fireAbstractGraphChangeEvent(e);
        afterSetProperties(properties);
    }

    public final String getValue(final String key) {
        return properties.get(key);
    }

    public final Set<String> getKeys() {
        return Collections.unmodifiableSet(properties.keySet());
    }

    private final synchronized void fireAbstractGraphChangeEvent(AbstractGraphChangeEvent e){
        for(AbstractGraphChangeListener l : listener){
            l.abstractGraphChangePerformed(e);
        }
    }

    protected void beforeAddNode(final AbstractGraphNode node) {
    }

    protected void afterAddNode(final AbstractGraphNode node) {
    }

    protected void beforeAddEdge(final AbstractGraphEdge edge) {
    }

    protected void afterAddEdge(final AbstractGraphEdge edge) {
    }

    protected void beforeRemoveNode(final AbstractGraphNode node) {
    }

    protected void afterRemoveNode(final AbstractGraphNode node) {
    }

    protected void beforeRemoveEdge(final AbstractGraphEdge edge) {
    }

    protected void afterRemoveEdge(final AbstractGraphEdge edge) {
    }

    protected void beforeAddProperty(final String key, final String value) {
    }

    protected void afterAddProperty(final String key, final String value) {
    }

    protected void beforeRemoveProperty(final String key) {
    }

    protected void afterRemoveProperty(final String key) {
    }

    protected void beforeSetProperties(final Map<String, String> properties) {
    }

    protected void afterSetProperties(final Map<String, String> properties) {
    }

    public final void abstractEdgeChangePerformed(AbstractGraphEdgeChangeEvent e) {
        fireAbstractEdgeChangeEvent(e);
    }

    private final synchronized void fireAbstractEdgeChangeEvent(AbstractGraphEdgeChangeEvent e) {
        for(AbstractEdgeChangeListener l : edgeListener){
            l.abstractEdgeChangePerformed(e);
        }
    }

    public final void abstractNodeChangePerformed(AbstractGraphNodeChangeEvent e) {
        fireAbstractNodeChangeEvent(e);
    }

    private final synchronized void fireAbstractNodeChangeEvent(AbstractGraphNodeChangeEvent e) {
        for(AbstractGraphNodeChangeListener l : nodeListener){
            l.abstractNodeChangePerformed(e);
        }
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.removePropertyChangeListener(listener);
    }

    public Lookup getLookup() {
        return lookup;
    }
}
