/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.gpdv.core.api.abstractgraph;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;

/**
 *
 * @author odi
 */
public class AbstractGraphEdge implements GraphPropertyProvider, Lookup.Provider{
    
    private final AbstractGraph graph;
    private final long id;
    private final String name;
    private AbstractGraphNode source;
    private AbstractGraphNode destination;
    private Map<String, String> properties;
    private final List<AbstractEdgeChangeListener> listener;
    private PropertyChangeSupport changeSupport;

    private boolean inserted;
    private final Lookup lookup;

    /**
     * Creates an implementation of AbstractGraphEdge
     * The created edge is automatically registered in the graph and linked to the source and destination nodes.
     * No additional linking needed!
     * @param id            The id of this edge, should be unique in graph, must not be null
     * @param name          The name of the edge, must not be null
     * @param graph         The AbstractGraph this edge belongs to, must not be null
     * @param source        The initial source, must belong to the specified graph, must not be null
     * @param destination   The initial destination, must belong to the specified graph, must not be null
     */
    public AbstractGraphEdge(final long id, final String name, final AbstractGraph graph, final AbstractGraphNode source, final AbstractGraphNode destination, Object object) throws IllegalArgumentException{
        if(name==null||graph==null||source==null||destination==null){
            throw new IllegalArgumentException();
        }
        this.id = id;
        this.name = name;
        properties = new HashMap<String, String>();
        listener = new ArrayList<AbstractEdgeChangeListener>();
        this.graph = graph;
        this.source = source;
        this.destination = destination;
        this.changeSupport = new PropertyChangeSupport(this);
        if(object!=null){
            this.lookup = Lookups.fixed(object);
        } else {
            this.lookup = Lookup.EMPTY;
        }
        /**
         * Exception is thrown if one of the following conditions is true:
         *      source and destination belong to different graphs
         *      source (and so destination) doesn't belong to specified graph
         *      source or destination currently are not in specified graph
         *      edge could not be inserted into the graph
         */
        if(source.getGraph()!=destination.getGraph()||source.getGraph()!=graph||!source.isInserted()||!destination.isInserted()||!graph.addEdge(this)){
            throw new IllegalArgumentException();
        }
    }

    final void inserted(){
        this.inserted = true;
    }

    final void removed(){
        this.inserted = false;
    }

    final boolean isInserted(){
        return inserted;
    }

    public final AbstractGraph getGraph() {
        return graph;
    }

    public final long getID() {
        return id;
    }

    public final String getName() {
        return name;
    }

    public final AbstractGraphNode getSource() {
        return source;
    }

    public final AbstractGraphNode getDestination() {
        return destination;
    }

    public final boolean setSource(final AbstractGraphNode node) {
        /*
         * returns false if one of the following conditions is true:
         *  node == null
         *  node == this.source
         *  node.graph != this.graph
         *  graph does not contain node
         */
        if(node==null||node==this.source||node.getGraph()!=this.graph||!node.isInserted()) return false;
        beforeSetSource(node);
        AbstractGraphEdgeChangeEvent e = new AbstractGraphEdgeChangeEvent(this, AbstractGraphEdgeChangeEvent.NEW_SOURCE, this.source, null);
        if(inserted) this.source.removeOutgoingEdge(this);
        this.source = node;
        if(inserted) this.source.addOutgoingEdge(this);
        this.fireAbstractEdgeChangeEvent(e);
        afterSetSource(node);
        return true;
    }

    /*
     * returns false if one of the following conditions is true:
     *  node == null
     *  node == this.destination
     *  node.graph != this.graph
     *  graph does not contain node
     */
    public final boolean setDestination(final AbstractGraphNode node) {
        if(node==null||this.destination==node||node.getGraph()!=this.graph||!node.isInserted()) return false;
        beforeSetDestination(node);
        AbstractGraphEdgeChangeEvent e = new AbstractGraphEdgeChangeEvent(this, AbstractGraphEdgeChangeEvent.NEW_DESTINATION, this.destination, null);
        if(inserted) this.destination.removeIncomingEdge(this);
        this.destination = node;
        if(inserted) this.destination.addIncomingEdge(this);
        this.fireAbstractEdgeChangeEvent(e);
        afterSetDestination(node);
        return true;
    }

    public final synchronized void addAbstractEdgeChangeListener(AbstractEdgeChangeListener listener) {
        if(!this.listener.contains(listener)) this.listener.add(listener);
    }

    public final synchronized void removeAbstractEdgeChangeListener(AbstractEdgeChangeListener listener) {
        this.listener.remove(listener);
    }

    public final Map<String, String> getProperties() {
        return Collections.unmodifiableMap(properties);
    }

    public final void addProperty(final String key, final String value) {
        if(key==null) return;
        beforeAddProperty(key, value);
        Map<String, String> map = new HashMap<String, String>();
        map.put(key, value);
        AbstractGraphEdgeChangeEvent e = new AbstractGraphEdgeChangeEvent(this, AbstractGraphEdgeChangeEvent.PROPERTY_ADDED, null, map);
        properties.put(key, value);
        this.fireAbstractEdgeChangeEvent(e);
        changeSupport.firePropertyChange(key, null, value);
        afterAddProperty(key, value);
    }

    public final void setProperties(final Map<String, String> properties) {
        beforeSetProperties(properties);
        Map<String, String> map = new HashMap<String, String>(this.properties);
        AbstractGraphEdgeChangeEvent e = new AbstractGraphEdgeChangeEvent(this, AbstractGraphEdgeChangeEvent.NEW_PROPERTIES, null, map);
        if(properties==null){
            for(String s : this.properties.keySet()){
                changeSupport.firePropertyChange(s, this.properties.get(s), null);
            }
            this.properties.clear();
        } else {
            this.properties = properties;
            for(String s : this.properties.keySet()){
                changeSupport.firePropertyChange(s, this.properties.get(s), null);
            }
        }
        this.fireAbstractEdgeChangeEvent(e);
        afterSetProperties(properties);
    }

    public final void removeProperty(final String key) {
        if(key==null||!properties.containsKey(key)) return;
        beforeRemoveProperty(key);
        Map<String, String> map = new HashMap<String, String>();
        map.put(key, properties.get(key));
        AbstractGraphEdgeChangeEvent e = new AbstractGraphEdgeChangeEvent(this, AbstractGraphEdgeChangeEvent.PROPERTY_REMOVED, null, map);
        String s = properties.get(key);
        this.properties.remove(key);
        changeSupport.firePropertyChange(key, s, null);
        this.fireAbstractEdgeChangeEvent(e);
        afterRemoveProperty(key);
    }

    public final String getValue(final String key) {
        return properties.get(key);
    }

    public final Set<String> getKeys() {
        return Collections.unmodifiableSet(properties.keySet());
    }

    protected void beforeSetSource(final AbstractGraphNode node) {
    }

    protected void afterSetSource(final AbstractGraphNode node) {
    }

    protected void beforeSetDestination(final AbstractGraphNode node) {
    }

    protected void afterSetDestination(final AbstractGraphNode node) {
    }

    protected void beforeAddProperty(final String key, final String value) {
    }

    protected void afterAddProperty(final String key, final String value) {
    }

    protected void beforeSetProperties(final Map<String, String> properties) {
    }

    protected void afterSetProperties(final Map<String, String> properties) {
    }

    protected void beforeRemoveProperty(final String key){
    }

    protected void afterRemoveProperty(final String key) {
    }

    protected final synchronized void fireAbstractEdgeChangeEvent(AbstractGraphEdgeChangeEvent e){
        if(!inserted) return;
        for(AbstractEdgeChangeListener l : listener){
            l.abstractEdgeChangePerformed(e);
        }
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.removePropertyChangeListener(listener);
    }

    public Lookup getLookup() {
        return lookup;
    }
}
