/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.gpdv.core.api.abstractgraph;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.openide.nodes.Children;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;

/**
 *
 * @author odi
 */
public class AbstractGraphNode implements GraphPropertyProvider, Lookup.Provider{
    private final AbstractGraph graph;
    private final long id;
    private final String name;
    private final List<AbstractGraphEdge> incomingEdges;
    private final List<AbstractGraphEdge> outgoingEdges;
    private Map<String, String> properties;
    private final List<AbstractGraphNodeChangeListener> listener;
    private final Lookup lookup;
    private PropertyChangeSupport changeSupport;

    private boolean inserted;

    /**
     * Creates an implementation of AbstractGraphNode
     * The node is automatically registered in the specified graph.
     * No additional linking needed!
     * @param nodeID
     * @param name
     * @param graph
     */
    public AbstractGraphNode(final long nodeID, final String name, final AbstractGraph graph, final Object object) throws IllegalArgumentException {
        if(name==null||graph==null) throw new IllegalArgumentException();
        this.id = nodeID;
        this.name = name;
        properties = new HashMap<String, String>();
        incomingEdges = new ArrayList<AbstractGraphEdge>();
        outgoingEdges = new ArrayList<AbstractGraphEdge>();
        listener = new ArrayList<AbstractGraphNodeChangeListener>();
        this.graph = graph;
        this.inserted = false;
        changeSupport = new PropertyChangeSupport(this);
        if(object!=null){
            this.lookup = Lookups.fixed(object);
        } else {
            this.lookup = Lookup.EMPTY;
        }
        if(!graph.addNode(this)) throw new IllegalArgumentException();
    }

    final void inserted(){
        this.inserted = true;
    }

    final void removed(){
        this.inserted = false;
    }

    final boolean isInserted(){
        return inserted;
    }

    public final AbstractGraph getGraph() {
        return graph;
    }

    public final long getID() {
        return id;
    }

    public final String getName() {
        return name;
    }

    public final Lookup getLookup(){
        return lookup;
    }

    public final List<AbstractGraphEdge> getIncomingEdges() {
        ArrayList<AbstractGraphEdge> l = new ArrayList<AbstractGraphEdge>();
        l.addAll(incomingEdges);
        return Collections.unmodifiableList(l);
    }

    public final List<AbstractGraphEdge> getOutgoingEdges() {
        ArrayList<AbstractGraphEdge> l = new ArrayList<AbstractGraphEdge>();
        l.addAll(outgoingEdges);
        return Collections.unmodifiableList(l);
    }

    final void addIncomingEdge(final AbstractGraphEdge edge) {
        if(edge==null||incomingEdges.contains(edge)) return;
        beforeAddIncomingEdge(edge);
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.INCOMING_EDGE_ADDED, edge, null);
        this.incomingEdges.add(edge);
        this.fireAbstractNodeChangeEvent(e);
        afterAddIncomingEdge(edge);
    }

    final void addOutgoingEdge(final AbstractGraphEdge edge) {
        if(edge==null||outgoingEdges.contains(edge)) return;
        beforeAddOutgoingEdge(edge);
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.OUTGOING_EDGE_ADDED, edge, null);
        this.outgoingEdges.add(edge);
        this.fireAbstractNodeChangeEvent(e);
        afterAddOutgoingEdge(edge);
    }

    final boolean removeIncomingEdge(final AbstractGraphEdge edge) {
        if(edge==null||!incomingEdges.contains(edge)) return false;
        beforeRemoveIncomingEdge(edge);
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.INCOMING_EDGE_REMOVED, edge, null);
        this.incomingEdges.remove(edge);
        this.fireAbstractNodeChangeEvent(e);
        afterRemoveIncomingEdge(edge);
        return true;
    }

    final boolean removeOutgoingEdge(final AbstractGraphEdge edge) {
        if(edge==null||!outgoingEdges.contains(edge)) return false;
        beforeRemoveOutgoingEdge(edge);
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.OUTGOING_EDGE_REMOVED, edge, null);
        this.outgoingEdges.remove(edge);
        this.fireAbstractNodeChangeEvent(e);
        afterRemoveOutgoingEdge(edge);
        return true;
    }

    public final synchronized void addAbstractNodeChangeListener(AbstractGraphNodeChangeListener listener) {
        this.listener.add(listener);
    }

    public final synchronized void removeAbstractNodeChangeListener(AbstractGraphNodeChangeListener listener) {
        this.listener.remove(listener);
    }

    public final Map<String, String> getProperties() {
        return Collections.unmodifiableMap(properties);
    }

    public final void addProperty(final String key, final String value) {
        if(key==null) return;
        beforeAddProperty(key, value);
        Map<String, String> map = new HashMap<String, String>();
        map.put(key, value);
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.PROPERTY_ADDED, null, map);
        properties.put(key, value);
        changeSupport.firePropertyChange(key, null, value);
        this.fireAbstractNodeChangeEvent(e);
        afterAddProperty(key, value);
    }

    public final void removeProperty(final String key) {
        if(key==null||!properties.containsKey(key)) return;
        beforeRemoveProperty(key);
        Map<String, String> map = new HashMap<String, String>();
        map.put(key, properties.get(key));
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.PROPERTY_REMOVED, null, map);
        String s = properties.get(key);
        this.properties.remove(key);
        changeSupport.firePropertyChange(key, s, null);
        this.fireAbstractNodeChangeEvent(e);
        afterRemoveProperty(key);
    }

    public final void setProperties(final Map<String, String> properties) {
        beforeSetProperties(properties);
        Map<String, String> map = new HashMap<String, String>(this.properties);
        AbstractGraphNodeChangeEvent e = new AbstractGraphNodeChangeEvent(this, AbstractGraphNodeChangeEvent.NEW_PROPERTIES, null, map);
        if(properties==null){
            for(String s : this.properties.keySet()){
                changeSupport.firePropertyChange(s, this.properties.get(s), null);
            }
            this.properties.clear();
        } else {
            this.properties = properties;
            for(String s : this.properties.keySet()){
                changeSupport.firePropertyChange(s, this.properties.get(s), null);
            }
        }
        this.fireAbstractNodeChangeEvent(e);
        afterSetProperties(properties);
    }

    public final String getValue(final String key) {
        return properties.get(key);
    }

    public final Set<String> getKeys() {
        return Collections.unmodifiableSet(properties.keySet());
    }

    protected final synchronized void fireAbstractNodeChangeEvent(AbstractGraphNodeChangeEvent e){
        if(!inserted) return;
        for(AbstractGraphNodeChangeListener l : listener){
            l.abstractNodeChangePerformed(e);
        }
    }

    protected void beforeAddIncomingEdge(final AbstractGraphEdge edge) {
    }

    protected void afterAddIncomingEdge(final AbstractGraphEdge edge) {
    }

    protected void beforeAddOutgoingEdge(final AbstractGraphEdge edge) {
    }

    protected void afterAddOutgoingEdge(final AbstractGraphEdge edge) {
    }

    protected void beforeRemoveIncomingEdge(final AbstractGraphEdge edge) {
    }

    protected void afterRemoveIncomingEdge(final AbstractGraphEdge edge) {
    }

    protected void beforeRemoveOutgoingEdge(final AbstractGraphEdge edge) {
    }

    protected void afterRemoveOutgoingEdge(final AbstractGraphEdge edge) {
    }

    protected void beforeAddProperty(final String key, final String value) {
    }

    protected void afterAddProperty(final String key, final String value) {
    }

    protected void beforeRemoveProperty(final String key) {
    }

    protected void afterRemoveProperty(final String key) {
    }

    protected void beforeSetProperties(Map<String, String> properties) {
    }

    protected void afterSetProperties(Map<String, String> properties) {
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        changeSupport.removePropertyChangeListener(listener);
    }
}
