/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.gpdv.core.api.designgraph;

import java.awt.Color;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.gpdv.core.api.abstractgraph.AbstractGraphEdge;
import org.gpdv.core.api.abstractgraph.AbstractGraphEdgeChangeEvent;
import org.gpdv.core.api.abstractgraph.AbstractEdgeChangeListener;
import org.gpdv.core.api.abstractgraph.AbstractGraph;
import org.gpdv.core.api.abstractgraph.AbstractGraphChangeEvent;
import org.gpdv.core.api.abstractgraph.AbstractGraphChangeListener;
import org.gpdv.core.api.abstractgraph.AbstractGraphNode;
import org.gpdv.core.api.DesignGraphFactory;
import org.gpdv.core.api.layoutgraph.LayoutGraphEdge;
import org.gpdv.core.api.layoutgraph.LayoutGraph;
import org.gpdv.core.api.layoutgraph.LayoutGraphChangeEvent;
import org.gpdv.core.api.layoutgraph.LayoutGraphChangeListener;
import org.gpdv.core.api.layoutgraph.LayoutGraphNode;

/**
 *
 * @author odi
 */
public class DesignGraph implements LayoutGraph, AbstractGraphChangeListener, AbstractEdgeChangeListener {

    private final AbstractGraph graph;
    private final List<DesignGraphNode> nodes;
    private final List<DesignGraphEdge> edges;
    private final List<LayoutGraphChangeListener> layoutListener;
    private final List<DesignGraphChangeListener> designListener;
    private final String name;
    private final long id;

    private final Map<AbstractGraphNode, DesignGraphNode> nodeMap;
    private final Map<AbstractGraphEdge, DesignGraphEdge> edgeMap;
    private final List<WeakReference<DesignGraphNode>> removedNodes;
    private final List<WeakReference<DesignGraphEdge>> removedEdges;

    public DesignGraph(final long graphID, final AbstractGraph graph) throws IllegalArgumentException{
        if(graph==null) throw new IllegalArgumentException();
        this.id = graphID;
        this.graph = graph;
        this.name = graph.getName();
        nodes = new ArrayList<DesignGraphNode>();
        edges = new ArrayList<DesignGraphEdge>();
        nodeMap = new HashMap<AbstractGraphNode, DesignGraphNode>();
        edgeMap = new HashMap<AbstractGraphEdge, DesignGraphEdge>();
        removedNodes = new ArrayList<WeakReference<DesignGraphNode>>();
        removedEdges = new ArrayList<WeakReference<DesignGraphEdge>>();
        designListener = new ArrayList<DesignGraphChangeListener>();
        layoutListener = new ArrayList<LayoutGraphChangeListener>();
        graph.addAbstractGraphChangeListener(this);
    }

    public final AbstractGraph getAbstractGraph(){
        return graph;
    }

    public final List<DesignGraphNode> getDesignNodes(){
        List<DesignGraphNode> l = new ArrayList<DesignGraphNode>();
        l.addAll(nodes);
        return Collections.unmodifiableList(l);
    }

    public final List<DesignGraphEdge> getDesignEdges(){
        List<DesignGraphEdge> l = new ArrayList<DesignGraphEdge>();
        l.addAll(edges);
        return Collections.unmodifiableList(l);
    }

    public final boolean addDesignNode(final DesignGraphNode node){
        if(node==null||node.isInserted()||node.getDesignGraph()!=this) return false;
        beforeAddDesignNode(node);
        DesignGraphChangeEvent e = new DesignGraphChangeEvent(this, DesignGraphChangeEvent.NODE_ADDED, node, null, null);
        LayoutGraphChangeEvent e1 = new LayoutGraphChangeEvent(this, LayoutGraphChangeEvent.NODE_ADDED, node, null);
        nodes.add(node);
        nodeMap.put(node.getAbstractNode(), node);
        node.inserted();
        this.fireDesignGraphChangeEvent(e);
        this.fireLayoutGraphChangeEvent(e1);
        afterAddDesignNode(node);
        return true;
    }

    public final boolean addDesignEdge(final DesignGraphEdge edge){
        if(edge==null||edge.isInserted()||edge.getDesignGraph()!=this) return false;
        if(!edge.getDestinationSlot().getDesignNode().isInserted()||!edge.getSourceSlot().getDesignNode().isInserted()) return false;
        beforeAddDesignEdge(edge);
        DesignGraphChangeEvent e = new DesignGraphChangeEvent(this, DesignGraphChangeEvent.EDGE_ADDED, null, edge, null);
        LayoutGraphChangeEvent e1 = new LayoutGraphChangeEvent(this, LayoutGraphChangeEvent.EDGE_ADDED, null, edge);
        edges.add(edge);
        edgeMap.put(edge.getAbstractEdge(), edge);
        edge.getSourceSlot().addDesignEdge(edge);
        edge.getDestinationSlot().addDesignEdge(edge);
        edge.inserted();
        this.fireDesignGraphChangeEvent(e);
        this.fireLayoutGraphChangeEvent(e1);
        afterAddDesignEdge(edge);
        return true;
    }

    public final boolean removeDesignNode(final DesignGraphNode node){
        if(node==null||node.getDesignGraph()!=this||!node.isInserted()) return false;
        beforeRemoveDesignNode(node);
        DesignGraphChangeEvent e = new DesignGraphChangeEvent(this, DesignGraphChangeEvent.NODE_REMOVED, node, null, null);
        LayoutGraphChangeEvent e1 = new LayoutGraphChangeEvent(this, LayoutGraphChangeEvent.NODE_REMOVED, node, null);
        for(DesignGraphSlot s : node.getIncomingSlots()){
            for(DesignGraphEdge edge : s.getDesignEdges()){
                this.removeDesignEdge(edge);
            }
        }
        for(DesignGraphSlot s : node.getOutgoingSlots()){
            for(DesignGraphEdge edge : s.getDesignEdges()){
                this.removeDesignEdge(edge);
            }
        }
        nodes.remove(node);
        nodeMap.remove(node.getAbstractNode());
        node.removed();
        removedNodes.add(new WeakReference<DesignGraphNode>(node));
        this.fireDesignGraphChangeEvent(e);
        this.fireLayoutGraphChangeEvent(e1);
        afterRemoveDesignNode(node);
        return true;
    }

    public final boolean removeDesignEdge(final DesignGraphEdge edge){
       if(edge==null||edge.getDesignGraph()!=this||!edge.isInserted()) return false;
        beforeRemoveDesignEdge(edge);
        DesignGraphChangeEvent e = new DesignGraphChangeEvent(this, DesignGraphChangeEvent.EDGE_REMOVED, null, edge, null);
        LayoutGraphChangeEvent e1 = new LayoutGraphChangeEvent(this, LayoutGraphChangeEvent.EDGE_REMOVED, null, edge);
        if(edge.getSourceSlot()!=null) edge.getSourceSlot().removeDesignEdge(edge);          // remove from source node
        if(edge.getDestinationSlot()!=null) edge.getDestinationSlot().removeDesignEdge(edge);     // remove from destination node
        edges.remove(edge);
        edgeMap.remove(edge.getAbstractEdge());
        edge.removed();
        removedEdges.add(new WeakReference<DesignGraphEdge>(edge));
        this.fireDesignGraphChangeEvent(e);
        this.fireLayoutGraphChangeEvent(e1);
        afterRemoveDesignEdge(edge);
        return true;
    }

    public synchronized final void addDesignGraphChangeListener(final DesignGraphChangeListener listener){
        this.designListener.add(listener);
    }

    public synchronized final void removeDesignGraphChangeListener(final DesignGraphChangeListener listener){
        this.designListener.remove(listener);
    }

    public final DesignGraph getDesignGraph() {
        return this;
    }

    public final String getName() {
        return name;
    }

    public final long getID() {
        return id;
    }

    public final List<LayoutGraphNode> getLayoutNodes() {
        List<LayoutGraphNode> l = new ArrayList<LayoutGraphNode>();
        l.addAll(nodes);
        return Collections.unmodifiableList(l);
    }

    public final List<LayoutGraphEdge> getLayoutEdges() {
        List<LayoutGraphEdge> l = new ArrayList<LayoutGraphEdge>();
        l.addAll(edges);
        return Collections.unmodifiableList(l);
    }

    public synchronized final void addLayoutGraphChangeListener(final LayoutGraphChangeListener listener) {
        this.layoutListener.add(listener);
    }

    public synchronized final void removeLayoutGraphChangeListener(final LayoutGraphChangeListener listener) {
        this.layoutListener.remove(listener);
    }

    private synchronized final void fireDesignGraphChangeEvent(final DesignGraphChangeEvent e) {
        for(DesignGraphChangeListener l : designListener){
            l.designGraphChangePerformed(e);
        }
    }

    private synchronized final void fireLayoutGraphChangeEvent(final LayoutGraphChangeEvent e1) {
        for(LayoutGraphChangeListener l : layoutListener){
            l.layoutGraphChangePerformed(e1);
        }
    }

    protected void beforeAddDesignNode(final DesignGraphNode node) {
    }

    protected void afterAddDesignNode(final DesignGraphNode node) {
    }

    protected void beforeAddDesignEdge(final DesignGraphEdge edge) {
    }

    protected void afterAddDesignEdge(final DesignGraphEdge edge) {
    }

    protected void beforeRemoveDesignNode(final DesignGraphNode node) {
    }

    protected void afterRemoveDesignNode(final DesignGraphNode node) {
    }

    protected void beforeRemoveDesignEdge(final DesignGraphEdge edge) {
    }

    protected void afterRemoveDesignEdge(final DesignGraphEdge edge) {
    }

    public synchronized final void abstractGraphChangePerformed(final AbstractGraphChangeEvent e) {
        if(e.getChangeString().equals(AbstractGraphChangeEvent.EDGE_ADDED)){
            // add edge
            // check if DesignGraphEdge is still available and insert if possible
            List<WeakReference<DesignGraphEdge>> list = new ArrayList<WeakReference<DesignGraphEdge>>(removedEdges);
            for(WeakReference<DesignGraphEdge> ref : list){
                DesignGraphEdge edge = ref.get();
                if(edge==null){
                    // reference is useless -> remove
                    removedEdges.remove(ref);
                } else if(edge.getAbstractEdge()==e.getChangedEdge()){
                    // reference points to old DesignGraphEdge representing AbstractGraphEdge
                    // check if AbstractGraphEdge has changed source or destination while not in graph
                    // inserted AbstractGraphEdge needs source and destination inserted into graph
                    // -> DesignNodes are also inserted
                    if(edge.getSourceSlot().getDesignNode().getAbstractNode()!=e.getChangedEdge().getSource()){
                        DesignGraphNode n = nodeMap.get(e.getChangedEdge().getSource());
                        if(n!=null) edge.setSource(findOutgoingSlot(n));
                    }
                    if(edge.getDestinationSlot().getDesignNode().getAbstractNode()!=e.getChangedEdge().getDestination()){
                        DesignGraphNode n = nodeMap.get(e.getChangedEdge().getDestination());
                        if(n!=null) edge.setDestination(findIncomingSlot(n));
                    }
                    removedEdges.remove(ref);
                    if(!addDesignEdge(edge)){
                        // UNEXPECTED BEHAVIOR
                        throw new RuntimeException();
                    }
                    return;
                }
            }
            DesignGraphEdge edge = createDesignEdge(e.getChangedEdge());
            if(edge==null||!edge.isInserted()){
                // deal with unexpected behavior
            }
        } else if(e.getChangeString().equals(AbstractGraphChangeEvent.EDGE_REMOVED)){
            // remove edge
            DesignGraphEdge edge = edgeMap.get(e.getChangedEdge());
            removeDesignEdge(edge);
        } else if(e.getChangeString().equals(AbstractGraphChangeEvent.NODE_ADDED)){
            // add node
            // check if DesignGraphNode is still available and insert if possible
            List<WeakReference<DesignGraphNode>> list = new ArrayList<WeakReference<DesignGraphNode>>(removedNodes);
            for(WeakReference<DesignGraphNode> ref : list){
                DesignGraphNode node = ref.get();
                if(node==null){
                    // reference is useless -> remove
                    removedNodes.remove(ref);
                } else if(node.getAbstractNode()==e.getChangedNode()){
                    // reference points to old DesignGraphNode representing AbstractGraphNode
                    removedNodes.remove(ref);
                    if(!addDesignNode(node)){
                        // UNEXPECTED BEHAVIOR
                        throw new RuntimeException();
                    }
                    return;
                }
            }
            DesignGraphNode node = createDesignNode(e.getChangedNode());
            if(node==null||!node.isInserted()){
                // deal with unexpected behaviour
            }
        } else if(e.getChangeString().equals(AbstractGraphChangeEvent.NODE_REMOVED)){
            // remove node
            DesignGraphNode node = nodeMap.get(e.getChangedNode());
            removeDesignNode(node);
        }
    }

    public synchronized final void abstractEdgeChangePerformed(final AbstractGraphEdgeChangeEvent e) {
        // AbstractEdgeChangeEvents are only fired while inserted
        // so the corresponding DesignGraphEdge must be inserted to
        if(e.getChangeString().equals(AbstractGraphEdgeChangeEvent.NEW_DESTINATION)){
            DesignGraphNode node = nodeMap.get(e.getEdge().getDestination());
            DesignGraphEdge edge = edgeMap.get(e.getEdge());
            if(edge==null){
                // deal with unexpected behavior
            }
            DesignGraphSlot s = findIncomingSlot(node);
            if(s==null){
                // deal with unexpected behavior
            } else {
                edge.setDestination(s);
            }
        } else if(e.getChangeString().equals(AbstractGraphEdgeChangeEvent.NEW_SOURCE)){
            DesignGraphNode node = nodeMap.get(e.getEdge().getSource());
            DesignGraphEdge edge = edgeMap.get(e.getEdge());
            if(edge==null){
                // deal with unexpected behavior
            }
            DesignGraphSlot s = findOutgoingSlot(node);
            if(s==null){
                // deal with unexpected behaviour
            } else {
                edge.setSource(s);
            }
        }
    }

    protected DesignGraphEdge createDesignEdge(AbstractGraphEdge edge) {
        DesignGraphNode sourceNode = nodeMap.get(edge.getSource());
        DesignGraphNode destinationNode = nodeMap.get(edge.getDestination());
        if(sourceNode.getOutgoingSlots().isEmpty()) DesignGraphFactory.getFactory().createOutgoingSlot(sourceNode);
        if(destinationNode.getIncomingSlots().isEmpty()) DesignGraphFactory.getFactory().createIncomingSlot(destinationNode);
        DesignGraphSlot sourceSlot = sourceNode.getOutgoingSlots().get(0);
        DesignGraphSlot destinationSlot = destinationNode.getIncomingSlots().get(0);
        return DesignGraphFactory.getFactory().createDesignEdge(edge, this, sourceSlot, destinationSlot, new Design(Color.BLACK,Color.RED, 0, 0));
    }

    protected DesignGraphNode createDesignNode(AbstractGraphNode node) {
        return DesignGraphFactory.getFactory().createDesignNode(node, this, new Design(Color.BLUE, Color.CYAN, 100, node.getName().length()*20));
    }

    protected DesignGraphSlot findIncomingSlot(DesignGraphNode node) {
        if(node.getIncomingSlots().isEmpty()) DesignGraphFactory.getFactory().createIncomingSlot(node);
        return node.getIncomingSlots().get(0);
    }

    protected DesignGraphSlot findOutgoingSlot(DesignGraphNode node) {
        if(node.getOutgoingSlots().isEmpty()) DesignGraphFactory.getFactory().createOutgoingSlot(node);
        return node.getOutgoingSlots().get(0);
    }
}
